@@ -23,53 +23,32 @@ import argparse
 import asyncio
 import collections
 import functools
-import warnings
 
 import requests
 import json
 import time
 from tqdm import tqdm
 from abc import abstractmethod
-from typing import Optional, Tuple, Dict, re, Any
+from typing import Optional, Tuple, Dict, Any
+import re
+import logging
 
 from idrac_ctl.shared import ApiRequestType, RedfishAction, ScheduleJobType
 from idrac_ctl.cmd_utils import save_if_needed
+from .cmd_exceptions import AuthenticationFailed
+from .cmd_exceptions import PostRequestFailed
+from .cmd_exceptions import MissingResource
+from .cmd_exceptions import UnexpectedResponse
+from .cmd_exceptions import ResourceNotFound
+from .cmd_exceptions import PatchRequestFailed
+from .cmd_exceptions import DeleteRequestFailed
+from .cmd_exceptions import UnsupportedAction
 
 """Each command encapsulate result in named tuple"""
 CommandResult = collections.namedtuple("cmd_result",
                                        ("data", "discovered", "extra"))
 
-
-class AuthenticationFailed(Exception):
-    pass
-
-
-class ResourceNotFound(Exception):
-    pass
-
-
-class MissingResource(Exception):
-    pass
-
-
-class UnexpectedResponse(Exception):
-    pass
-
-
-class PatchRequestFailed(Exception):
-    pass
-
-
-class PostRequestFailed(Exception):
-    pass
-
-
-class DeleteRequestFailed(Exception):
-    pass
-
-
-class UnsupportedAction(Exception):
-    pass
+module_logger = logging.getLogger('idrac_ctl.idrac_manager')
 
 
 class IDracManager:
@@ -98,12 +77,18 @@ class IDracManager:
         self._password = idrac_password
         self._is_verify_cert = insecure
         self._x_auth = x_auth
+
+        self.logger = logging.getLogger(__name__)
+
         self.content_type = {'Content-Type': 'application/json; charset=utf-8'}
         self.json_content_type = {'Content-Type': 'application/json; charset=utf-8'}
 
         # run time
         self.action_targets = None
         self.api_endpoints = None
+        self._post_success_responses = [200, 201, 202, 203, 204]
+        self._patch_success_responses = [200, 201, 202, 203, 204]
+        self._delete_success_responses = [200, 201, 202, 203, 204]
 
     @property
     def idrac_ip(self):
@@ -157,12 +142,15 @@ class IDracManager:
     @classmethod
     def invoke(cls, api_call: ApiRequestType, name: str, **kwargs) -> CommandResult:
         """Main interface uses to invoke a command.
-        :param api_call:
-        :param name:
-        :param kwargs:
+        :param api_call: api request type is enum for each cmd.
+        :param name: a name is key for a given api request type.
+                      So we can register under same type sub-commands.
+        :param kwargs: args passed to command.
         :return:
         """
         z = cls._registry[api_call]
+        if name not in z:
+            raise UnsupportedAction(f"Unknown {name} command.")
         disp = z[name]
         _idrac_ip = kwargs.pop("idrac_ip")
         _username = kwargs.pop("username")
@@ -175,13 +163,15 @@ class IDracManager:
 
     async def async_invoke(cls, api_call: ApiRequestType, name: str, **kwargs) -> CommandResult:
         """Main interface uses to invoke a command.
-        :param api_call:
+        :param api_call: api request type is enum for each cmd.
         :param name:
         :param kwargs:
         :return:
         """
         z = cls._registry[api_call]
         disp = z[name]
+        if name not in z:
+            raise UnsupportedAction(f"Unknown {name} command.")
         _idrac_ip = kwargs.pop("idrac_ip")
         _username = kwargs.pop("username")
         _password = kwargs.pop("password")
@@ -227,7 +217,7 @@ class IDracManager:
         await self.async_default_error_handler(await response)
         return await response
 
-    def api_get_call(self, r, hdr: Dict):
+    def api_get_call(self, r, hdr: Dict) -> requests.models.Response:
         """Make api request either with x-auth authentication header or idrac_ctl.
         :param r:  request
         :param hdr: http header dict that will append to HTTP/HTTPS request.
@@ -316,9 +306,8 @@ class IDracManager:
                                 pbar.update(n=inc)
                             time.sleep(sleep_time)
                 else:
-                    print("unexpected status code", resp.status_code)
+                    self.logger.error("unexpected status code", resp.status_code)
                     time.sleep(sleep_time)
-                    print("Unknown status code")
 
         return resp_data
 
@@ -392,7 +381,6 @@ class IDracManager:
         else:
             json_raw = json.dumps(json_data,
                                   sort_keys=sort, indent=indents)
-
         print(json_raw)
 
     @staticmethod
@@ -518,7 +506,7 @@ class IDracManager:
                                  if attr_filter.lower() in attr.lower())
         return json_data
 
-    def api_post_call(self, req: str, payload: str, hdr: dict):
+    def api_post_call(self, req: str, payload: str, hdr: dict) -> requests.models.Response:
         """Make api post request.
         :param req: path to a path request
         :param payload:  json payload
@@ -716,7 +704,8 @@ class IDracManager:
         return err_msg
 
     @staticmethod
-    def default_patch_success(cls, response: requests.models.Response,
+    def default_patch_success(cls,
+                              response: requests.models.Response,
                               expected: Optional[int] = 200) -> bool:
         """Default HTTP patch success handler
         Default handler to check patch request respond.
@@ -730,7 +719,8 @@ class IDracManager:
         if response.status_code == expected:
             return True
 
-        if response.status_code == 200 or response.status_code == 202:
+        if response.status_code == 200 \
+                or response.status_code == 202 or response.status_code == 204:
             return True
         else:
             err_msg = IDracManager.parse_error(response)
@@ -843,12 +833,13 @@ class IDracManager:
         :return: CommandResult and if filename provide will save to a file.
         """
         if verbose:
-            print(f"cmd args"
-                  f"data_type: {data_type} "
-                  f"resource:{resource} "
-                  f"do_async:{do_async} "
-                  f"filename:{filename}")
-            print(f"the rest of args: {kwargs}")
+            self.logger.info(
+                f"cmd args"
+                f"data_type: {data_type} "
+                f"resource:{resource} "
+                f"do_async:{do_async} "
+                f"filename:{filename}")
+            self.logger.info(f"the rest of args: {kwargs}")
 
         headers = {}
         if data_type == "json":
@@ -904,7 +895,7 @@ class IDracManager:
                 loop = asyncio.get_event_loop()
                 ok, response = loop.run_until_complete(self.api_async_patch_until_complete(r, json.dumps(pd), headers))
         except PatchRequestFailed as pf:
-            print("Error:", pf)
+            self.logger.critical(pf, exc_info=True)
             pass
 
         return CommandResult(self.api_success_msg(ok), None, response)
@@ -914,13 +905,15 @@ class IDracManager:
                   payload: Optional[dict] = None,
                   do_async: Optional[bool] = False,
                   data_type: Optional[str] = "json",
-                  expected_status: Optional[int] = 200) -> CommandResult:
-        """Base http post request..
-        :param resource:
-        :param payload:
+                  expected_status: Optional[int] = 20,
+                  verbose: Optional[bool] = False) -> CommandResult:
+        """Base http post request
+        :param resource: a remote resource
+        :param payload: a json payload
         :param do_async:
         :param data_type:
         :param expected_status:
+        :param verbose: enables verbose output
         :return:
         """
         headers = {}
@@ -938,16 +931,18 @@ class IDracManager:
             r = f"https://{self.idrac_ip}{resource}"
             if not do_async:
                 response = self.api_post_call(r, json.dumps(pd), headers)
+                if verbose:
+                    print(f"received status code {response.status_code}")
+                    print(f"received status code {response.headers}")
+                    print(f"received {response.raw}")
                 ok = self.default_post_success(self, response, expected=expected_status)
             else:
                 loop = asyncio.get_event_loop()
                 ok, response = loop.run_until_complete(
                     self.async_post_until_complete(r, json.dumps(pd), headers)
                 )
-
         except PostRequestFailed as pf:
-            print("Error:", pf)
-            pass
+            self.logger.critical(pf, exc_info=True)
 
         return CommandResult(self.api_success_msg(ok), None, response)
 
@@ -973,15 +968,16 @@ class IDracManager:
         if isinstance(cmd_result.data, dict) and 'PowerState' in cmd_result.data:
             pd_state = cmd_result.data[power_state_attr]
             if pd_state == 'On':
-                cmd_result = self.sync_invoke(ApiRequestType.RebootHost,
-                                              "reboot",
-                                              reset_type=default_reboot_type)
+                cmd_result = self.sync_invoke(
+                    ApiRequestType.RebootHost, "reboot",
+                    reset_type=default_reboot_type
+                )
                 if 'Status' in cmd_result.data:
                     result_data.update({"Reboot": cmd_result.data['Status']})
             else:
-                warnings.warn(f"Can't reboot a host , chassis power state {pd_state}")
+                self.logger.info(f"Can't reboot a host, chassis power state in {pd_state} state.")
         else:
-            warnings.warn(f"Failed fetch chassis power state")
+            self.logger.info(f"Failed to fetch chassis power state")
 
         return result_data
 
@@ -1044,30 +1040,35 @@ class IDracManager:
 
     @staticmethod
     def job_id_from_respond(response: requests.models.Response) -> Any | None:
-        """parse job id from a respond.
+        """Parses job id from a HTTP respond.
         :param response:
         :return:
         """
-        response_dict = str(response.__dict__)
         try:
-            job_id = re.search("JID_.+?,", response_dict).group()
-            return job_id
-        except Exception as err:
-            warnings.warn(str(err))
+            if response is not None:
+                response_dict = str(response.__dict__)
+                if response_dict is not None and len(response_dict) > 0:
+                    job_id = re.search("JID_.+?,", response_dict)
+                    if job_id is not None:
+                        job_id = job_id.group(0)
+                    return job_id
+        except AttributeError as _:
             pass
 
         return None
 
     @staticmethod
     def job_id_from_header(response: requests.models.Response) -> str:
-        """Returns job id from response header.
-        :param response:
-        :return:
+        """Returns job id from the response header.
+        :param response: a response that should have job
+        id information in the header.
+        :return: job id
         :raise UnexpectedResponse if header not present.
         """
         resp_hdr = response.headers
         if 'Location' not in resp_hdr:
-            raise UnexpectedResponse("rest api failed.")
+            raise UnexpectedResponse("There is no location in the "
+                                     "response header. (not all api create job id)")
 
         location = response.headers['Location']
         job_id = location.split("/")[-1]
@@ -1077,7 +1078,7 @@ class IDracManager:
     def schedule_job(reboot_type: ScheduleJobType,
                      start_time: Optional[str],
                      duration_time: Optional[int]) -> dict:
-        """
+        """Schedule a job.
         :param reboot_type: reboot types.
         :param start_time: start time for a job
         :param duration_time: duration for a job
@@ -1108,11 +1109,14 @@ class IDracManager:
         return pd
 
     def parse_task_id(self, data):
-        """
+        """Parse input data and try to get job id from the header or response.
         :param data:
         :return:
         """
         # get response from extra
+        if data is None:
+            return {}
+
         if hasattr(data, "extra"):
             resp = data.extra
         elif isinstance(data, requests.models.Response):
@@ -1120,18 +1124,25 @@ class IDracManager:
         else:
             raise ValueError("Unknown data type.")
 
+        if resp is None:
+            return {}
+
         job_id = None
         try:
             job_id = self.job_id_from_header(resp)
+            logging.debug(f"idrac api returned {job_id} in the header.")
         except UnexpectedResponse as ur:
+            logging.debug(ur, exc_info=False)
+
+        try:
+            # try to get from the response , it an optional check.
+            if job_id is None:
+                job_id = self.job_id_from_respond(resp)
+                logging.debug(f"idrac api returned {job_id} in the header.")
+        except UnexpectedResponse as _:
             pass
 
-        # try to get from response
-        if job_id is None:
-            job_id = self.job_id_from_respond(resp)
-
         if job_id is not None:
-            print(f"job_id {job_id}")
             data = self.fetch_job(job_id)
             return data
 
